@@ -10,6 +10,8 @@ use std::iter::Peekable;
 #[derive(Debug, Clone, Eq, Hash, PartialEq)]
 pub enum OpCode {
     // VM Primitives
+    Jmp,
+    Jnt,
     Mov,
     MovImmediate,
     Push,
@@ -22,31 +24,13 @@ pub enum OpCode {
     ClosureAcc,
     Enter,
     Ret,
-
-    //Lists
-    Car,
-    Cdr,
-    Cons,
+    TCallAcc,
+    VarArg,
 
     //Numbers
     Add,
     Mul,
     Sub,
-
-    // Predicates
-    Eq,
-    IsBoolean,
-    IsChar,
-    IsList,
-    IsNumber,
-    IsNull,
-    IsPair,
-    IsPort,
-    IsProcedure,
-    IsString,
-    IsSymbol,
-    IsVector,
-    Not,
 }
 
 #[derive(Debug, Clone, Eq, PartialEq)]
@@ -114,32 +98,20 @@ fn schema() -> &'static HashMap<OpCode, Schema> {
             (OpCode::Add, Schema::new("ADD", vec![])),
             (OpCode::CallAcc, Schema::new("CALL", vec![Operand::Acc])),
             (OpCode::ClosureAcc, Schema::new("CLOSURE", vec![Operand::Acc])),
-            (OpCode::Car, Schema::new("CAR", vec![])),
-            (OpCode::Cdr, Schema::new("CDR", vec![])),
-            (OpCode::Cons, Schema::new("CONS", vec![])),
             (OpCode::Enter, Schema::new("ENTER", vec![])),
-            (OpCode::Eq, Schema::new("EQ", vec![])),
             (OpCode::Halt, Schema::new("HALT", vec![])),
-            (OpCode::IsBoolean, Schema::new("BOOLEAN?", vec![Operand::Acc])),
-            (OpCode::IsChar, Schema::new("CHAR?", vec![Operand::Acc])),
-            (OpCode::IsList, Schema::new("LIST?", vec![Operand::Acc])),
-            (OpCode::IsNumber, Schema::new("NUMBER?", vec![Operand::Acc])),
-            (OpCode::IsNull, Schema::new("NULL?", vec![Operand::Acc])),
-            (OpCode::IsPair, Schema::new("PAIR?", vec![Operand::Acc])),
-            (OpCode::IsPort, Schema::new("PORT?", vec![Operand::Acc])),
-            (OpCode::IsProcedure, Schema::new("PROCEDURE?", vec![Operand::Acc])),
-            (OpCode::IsString, Schema::new("STRING?", vec![Operand::Acc])),
-            (OpCode::IsSymbol, Schema::new("SYMBOL?", vec![Operand::Acc])),
-            (OpCode::IsVector, Schema::new("VECTOR?", vec![Operand::Acc])),
+            (OpCode::Jmp, Schema::new("JMP", vec![Operand::Immediate])),
+            (OpCode::Jnt, Schema::new("JNT", vec![Operand::Immediate])),
             (OpCode::Mov, Schema::new("MOV", vec![Operand::LoadReference, Operand::StoreReference])),
             (OpCode::MovImmediate, Schema::new("MOV", vec![Operand::Immediate, Operand::StoreReference])),
             (OpCode::Mul, Schema::new("MUL", vec![])),
-            (OpCode::Not, Schema::new("NOT", vec![Operand::Acc])),
             (OpCode::Push, Schema::new("PUSH", vec![Operand::LoadReference])),
             (OpCode::PushImmediate, Schema::new("PUSH", vec![Operand::Immediate])),
             (OpCode::PushAcc, Schema::new("PUSH", vec![Operand::Acc])),
             (OpCode::Ret, Schema::new("RET", vec![])),
-            (OpCode::Sub, Schema::new("SUB", vec![]))
+            (OpCode::Sub, Schema::new("SUB", vec![])),
+            (OpCode::TCallAcc, Schema::new("TCALL", vec![Operand::Acc])),
+            (OpCode::VarArg, Schema::new("VARARG", vec![]))
         ]);
     }
     &SCHEMA
@@ -269,7 +241,7 @@ mod tests {
                 ]))
                 .as_ptr()
                 .unwrap();
-
+            vm.ip.1 = 0;
             assert!(vm.run().is_ok());
             assert_eq!(vm.acc, ptr);
         }
@@ -287,6 +259,7 @@ mod tests {
                 ]))
                 .as_ptr()
                 .unwrap();
+            vm.ip.1 = 0;
             assert!(vm.run().is_ok());
             assert_eq!(vm.acc, VCell::Bool(true));
         }
