@@ -2,10 +2,9 @@ use crate::cell::Cell;
 use crate::vm::environment::{BindingSource, EnvironmentMap, LexicalEnvironment};
 use crate::vm::lambda::Lambda;
 use crate::vm::opcode::OpCode;
-use crate::vm::vcell::VCell;
+use crate::vm::vcell::{BuiltInProc, VCell};
 use crate::vm::Error::{
-    ExpectedPairButFound, InvalidArgs, InvalidBytecode, InvalidNumArgs, InvalidProcedure,
-    VariableNotBound,
+    InvalidArgs, InvalidBytecode, InvalidNumArgs, InvalidProcedure, VariableNotBound,
 };
 use crate::vm::{Error, Vm};
 use log::trace;
@@ -17,13 +16,20 @@ impl Vm {
     /// Run the virtual machine until it encounters a HALT instruction,
     /// and return the value contained within the ACC register as a Cell
     pub fn run(&mut self) -> Result<Cell, Error> {
+        let mut cycles = 0;
+        self.stack.clear();
         loop {
+            cycles += 1;
+            if cycles % 256 == 0 {
+                self.run_gc();
+            }
             match self.run_one() {
                 Ok(true) => break,
                 Ok(false) => continue,
                 Err(e) => return Err(e),
             }
         }
+        trace!("cycles: {}", cycles);
         let cell = self.heap.get_as_cell(&self.acc);
         self.run_gc();
         Ok(cell)
@@ -43,6 +49,15 @@ impl Vm {
             // The instructions in this section are abstract virtual machine instructions
             // that provide support for the scheme VM. Operations such as stack manipulation,
             // loading and storing via MOV, and HALT.
+            OpCode::Jmp => {
+                self.ip.1 = self.read_operand()?.as_ptr()?;
+            }
+            OpCode::Jnt => {
+                let offset = self.read_operand()?.as_ptr()?;
+                if let VCell::Bool(false) = self.heap.get(&self.acc) {
+                    self.ip.1 = offset;
+                }
+            }
             OpCode::Mov => {
                 let vcell = self.load_operand()?;
                 self.store_operand(vcell)?;
@@ -91,6 +106,10 @@ impl Vm {
                 let lambda = match self.heap.get(&self.acc) {
                     VCell::Closure(lambda, _) => lambda,
                     VCell::Lambda(_) => self.acc.as_ptr()?,
+                    VCell::BuiltInProc(BuiltInProc(func)) => {
+                        func(self)?;
+                        return Ok(false);
+                    }
                     other => {
                         return Err(InvalidProcedure(self.heap.get_as_cell(&other).to_string()));
                     }
@@ -101,6 +120,52 @@ impl Vm {
                 self.ip.0 = lambda;
                 self.ip.1 = 0;
             }
+            OpCode::TCallAcc => {
+                let lambda = match self.heap.get(&self.acc) {
+                    VCell::Closure(lambda, _) => lambda,
+                    VCell::Lambda(_) => self.acc.as_ptr()?,
+                    VCell::BuiltInProc(BuiltInProc(func)) => {
+                        func(self)?;
+                        return Ok(false);
+                    }
+                    other => {
+                        return Err(InvalidProcedure(self.heap.get_as_cell(&other).to_string()));
+                    }
+                };
+
+                let argc = self.stack.get_offset(0)?.as_fixed_num()? as usize;
+                let frame_argc = self.stack.get(self.bp + 1)?.as_fixed_num()? as usize;
+
+                if argc == frame_argc {
+                    let saved_bp = self.stack.get(self.bp + 4)?.clone();
+                    for it in 0..argc {
+                        let val = self.stack.get_offset(-1_i64 - it as i64)?.clone();
+                        *self.stack.get_mut(self.bp - it)? = val;
+                    }
+                    *self.stack.get_sp_mut() = self.bp + 3;
+                    self.bp = saved_bp.as_bp()?;
+                    self.ip.0 = lambda;
+                    self.ip.1 = 0;
+                } else {
+                    let saved_sp = self.stack.get_sp();
+                    let saved_ep = self.stack.get(self.bp + 2)?.clone();
+                    let saved_ip = self.stack.get(self.bp + 3)?.clone();
+                    let saved_bp = self.stack.get(self.bp + 4)?.clone();
+
+                    *self.stack.get_sp_mut() = self.bp - frame_argc;
+                    for it in (0..argc).rev() {
+                        let val = self.stack.get(saved_sp - it - 1)?.clone();
+                        self.stack.push(val);
+                    }
+
+                    self.stack.push(VCell::FixedNum(argc as i64));
+                    self.stack.push(saved_ep);
+                    self.stack.push(saved_ip);
+                    self.bp = saved_bp.as_bp()?;
+                    self.ip.0 = lambda;
+                    self.ip.1 = 0;
+                }
+            }
             OpCode::Enter => {
                 let (lambda, lexical_env) = match self.heap.get(&self.acc) {
                     VCell::Closure(lambda, lexical_env) => (lambda, lexical_env),
@@ -127,38 +192,45 @@ impl Vm {
                 self.ip = self.stack.get(self.bp + 3)?.as_ip()?;
                 self.bp = self.stack.get(self.bp + 4)?.as_bp()?;
             }
-            // Lists
-            //
-            // The opcodes in this section provide primitive instructions for manipulating lists,
-            // such as the primitives for `car`, `cdr` and `cons`
-            //
-            OpCode::Car => {
-                let arg = self.heap.get(&self.acc);
-                match car(&arg) {
-                    Ok(vcell) => self.acc = vcell,
-                    Err(_) => {
-                        return Err(ExpectedPairButFound(
-                            self.heap.get_as_cell(&arg).to_string(),
-                        ));
-                    }
+            OpCode::VarArg => {
+                // VARARG converts the optional arguments of a vararg procedure into a list.
+                // This requires popping off every optional argument, forming a list on the heap,
+                // and placing the list at the top of the argument stack.
+                let req_argc = self.lambda().args.len() - 1;
+                let argc = self.stack.get_offset(-2)?.as_fixed_num()? as usize;
+                if argc < req_argc {
+                    return Err(InvalidNumArgs("procedure".into()));
                 }
-            }
-            OpCode::Cdr => {
-                let arg = self.heap.get(&self.acc);
-                match cdr(&arg) {
-                    Ok(vcell) => self.acc = vcell,
-                    Err(_) => {
-                        return Err(ExpectedPairButFound(
-                            self.heap.get_as_cell(&arg).to_string(),
-                        ));
+
+                // If there's exactly one vararg, then we can convert it in place
+                if argc == req_argc + 1 {
+                    let arg = self.stack.get_offset(-3)?.clone();
+                    let nil = self.heap.put(VCell::Nil);
+                    *self.stack.get_offset_mut(-3)? =
+                        self.heap.put(VCell::Pair(arg.as_ptr()?, nil.as_ptr()?));
+                } else {
+                    // Save the frame data that CALL put on the stack
+                    let saved_ep = self.stack.pop()?.clone();
+                    let saved_ip = self.stack.pop()?.clone();
+                    let _ = self.stack.pop()?.clone();
+
+                    // Pop each optional arg into a list
+                    let varargc = argc - req_argc;
+                    let mut varargs = self.heap.put(VCell::Nil).as_ptr()?;
+                    for _ in 0..varargc {
+                        let arg = self.stack.pop()?.clone();
+                        let pair = VCell::Pair(arg.as_ptr()?, varargs);
+                        varargs = self.heap.put(pair).as_ptr()?;
                     }
+
+                    // Push the list on the stack, a new argc, and restore the caller's
+                    // frame
+                    self.stack.push(VCell::ptr(varargs));
+                    self.stack.push(VCell::FixedNum((req_argc + 1) as i64));
+                    self.stack.push(saved_ip);
+                    self.stack.push(saved_ep);
                 }
             }
-            OpCode::Cons => {
-                let car = self.acc.clone().as_ptr()?;
-                let cdr = self.pop_stack()?.as_ptr()?;
-                self.acc = self.heap.put(VCell::Pair(car, cdr));
-            }
             // Numbers
             //
             // The opcodes in this section provide primitives for monipulating numbers,
@@ -195,66 +267,6 @@ impl Vm {
                     _ => return Err(InvalidBytecode),
                 };
             }
-            // Predicates
-            //
-            // These opcodes provide predicates for a single object, or for equievalence of
-            // objects. Predicates always evaluate to #t or #f
-            OpCode::Eq => {
-                let arg = self.pop_stack()?;
-                self.acc = self.heap.put(self.acc == arg);
-            }
-            OpCode::IsBoolean => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_boolean());
-            }
-            OpCode::IsChar => {
-                self.acc = self.heap.put(false);
-            }
-            OpCode::IsList => {
-                let mut lat = self.heap.get(&self.acc);
-                loop {
-                    if !lat.is_pair() {
-                        self.acc = self.heap.put(lat.is_nil());
-                        break;
-                    }
-                    lat = self.heap.get(&lat.as_cdr()?);
-                }
-            }
-            OpCode::IsNull => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_nil());
-            }
-            OpCode::IsNumber => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_number());
-            }
-            OpCode::IsPair => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_pair());
-            }
-            OpCode::IsPort => {
-                self.acc = self.heap.put(false);
-            }
-            OpCode::IsProcedure => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_procedure());
-            }
-            OpCode::IsString => {
-                self.acc = self.heap.put(false);
-            }
-            OpCode::IsSymbol => {
-                let arg = self.heap.get(&self.acc);
-                self.acc = self.heap.put(arg.is_symbol());
-            }
-            OpCode::IsVector => {
-                self.acc = self.heap.put(false);
-            }
-            OpCode::Not => {
-                self.acc = match self.heap.get(&self.acc) {
-                    VCell::Bool(val) => self.heap.put(!val),
-                    _ => self.heap.put(false),
-                };
-            }
         }
 
         Ok(false)
@@ -436,6 +448,8 @@ impl Vm {
             .filter_map(|it| it.as_ptr().ok())
             .for_each(|it| self.heap.mark(it));
 
+        self.stack.iter().for_each(|it| self.heap.mark_vcell(it));
+        self.heap.mark_vcell(&self.acc);
         self.heap.mark(self.ip.0);
         self.heap.sweep();
     }
@@ -495,7 +509,8 @@ impl Vm {
                 .unwrap()
             ),
             format!(
-                "{} %sp[{}] {} $ep[{}]",
+                "%acc={} {} %sp[{}] {} $ep[{}]",
+                self.acc,
                 VCell::InstructionPointer(self.ip.0, self.ip.1),
                 VCell::Ptr(self.stack.get_sp()),
                 VCell::BasePointer(self.bp),
@@ -505,15 +520,5 @@ impl Vm {
     }
 }
 
-fn car<T: AsRef<VCell>>(vcell: T) -> Result<VCell, Error> {
-    let vcell = vcell.as_ref();
-    vcell.as_car()
-}
-
-fn cdr<T: AsRef<VCell>>(vcell: T) -> Result<VCell, Error> {
-    let vcell = vcell.as_ref();
-    vcell.as_cdr()
-}
-
 #[cfg(test)]
 mod tests {}
